"""
    compute_rhs_weak_form!(rhs, f, nf, mesh::Mesh1d{DGElement})
    compute_rhs_weak_form!(rhs, f, s, nf, mesh::Mesh1d{DGElement})

Compute the `rhs` of the weak DG formulation using
the (bulk) flux `f`, the numerical flux `nf` and sources `s` for a `mesh`.
"""
function compute_rhs_weak_form!(rhs, f, nf, mesh::Mesh1d{DGElement})
  @unpack element = mesh
  @unpack invM, MDM, Ml_lhs, Ml_rhs, Npts = element
  # @unpack K = mesh
  K = mesh.tree.dims[1]
  @unpack invdetJ = get_static_variables(mesh.cache)
  shape       = layout(mesh)
  mat_rhs     = vreshape(rhs,     shape)
  mat_f       = vreshape(f,       shape)
  mat_invdetJ = vreshape(invdetJ, shape)
  mul!(mat_rhs, MDM, mat_f)
  nf_lhs       = view(nf, 1:2:2*K)
  nf_rhs       = view(nf, 2:2:2*K)
  @turbo @. begin # factor x10
    mat_rhs  -= (nf_rhs' * Ml_rhs + #= minus contained in normal vector =# nf_lhs' * Ml_lhs)
    mat_rhs *= mat_invdetJ
  end
  return
end
function compute_rhs_weak_form!(rhs, f, s, nf, mesh::Mesh1d{DGElement})
  compute_rhs_weak_form!(rhs, f, nf, mesh)
  rhs .+= s
  return
end
function compute_rhs_weak_form!(rhs, f, rhs_modal, f_modal, nf, mesh::Mesh1d{DGElement},
                                element::DGElement, mask::AbstractVector, workspace)
  @unpack Npts = mesh.element
  @unpack invM, MDM, Ml_lhs, Ml_rhs = element
  # @unpack K = mesh
  K = mesh.tree.dims[1]
  @unpack invdetJ = get_static_variables(mesh)
  tmp_rhs, tmp_f = workspace.buf1, workspace.buf2
  invJ = invdetJ[1]
  nf_lhs = view(nf, 1:2:2*K)
  nf_rhs = view(nf, 2:2:2*K)
  for k in 1:K
    mask[k] > 0.0 || continue
    cidxs = cellindices(mesh, k)
    rng_modal = 1:Npts-1
    v_f       = view(f, cidxs)
    v_f_modal = view(f_modal, cidxs)
    @views tmp_f[1 .+ rng_modal] .= v_f_modal[rng_modal]
    tmp_f[1] = v_f[1]; tmp_f[end] = v_f[end]
    mul!(tmp_rhs, MDM, tmp_f)
    @. tmp_rhs -= (nf_rhs[k] * Ml_rhs + #= minus contained in normal vector =# nf_lhs[k] * Ml_lhs)
    @. tmp_rhs *= invJ
    v_rhs       = view(rhs, cidxs)
    v_rhs_modal = view(rhs_modal, cidxs)
    @views v_rhs_modal[rng_modal] .= tmp_rhs[1 .+ rng_modal]
    v_rhs[1] = tmp_rhs[1]; v_rhs[end] = tmp_rhs[end]
  end
  return
end
function compute_rhs_weak_form!(rhs, f, s, rhs_modal, f_modal, s_modal, nf, mesh::Mesh1d{DGElement},
                                bspline::DGElement, mask::AbstractVector, workspace)
  compute_rhs_weak_form!(rhs, f, rhs_modal, f_modal, nf, mesh, bspline, mask, workspace)
  @unpack Npts = mesh.element
  for (k,(v_rhs,v_s)) in enumerate(zip(eachcell(mesh,rhs),eachcell(mesh,s)))
    mask[k] > 0.0 || continue
    cidxs = cellindices(mesh, k)
    v_rhs_modal = view(rhs_modal, cidxs)
    v_s_modal   = view(s_modal, cidxs)
    rng_modal = 1:Npts-1
    v_rhs[1] += v_s[1]; v_rhs[end] += v_s[end]
  end
  return
end


"""
    compute_rhs_weak_form!(rhs, fx, fy, nf, mesh::Mesh2d{DGElement})
    compute_rhs_weak_form!(rhs, fx, fy, s, nf, mesh::Mesh2d{DGElement})

Compute the `rhs` of the weak DG formulation using
the (bulk) flux `fx, fy`, the numerical flux `nf` and sources `s` for a `mesh`.
"""
function compute_rhs_weak_form!(rhs, fx, fy, nf, mesh::Mesh2d{DGElement})
  @unpack element = mesh
  @unpack dxdX, dydY = get_static_variables(mesh.cache)
  @unpack invM, MDM, Ml_lhs, Ml_rhs, Npts = element
  Ml_down, Ml_up = Ml_lhs, Ml_rhs
  Nx, Ny = Npts, Npts
  Ncs = n_cells(mesh.tree)
  @inbounds for kk in 1:Ncs

    # bulk
    idxs_bulk = flatcellindices(mesh, kk)
    v_rhs     = rview(rhs,  idxs_bulk, (Nx,Ny))
    v_fx      = rview(fx,   idxs_bulk, (Nx,Ny))
    v_fy      = rview(fy,   idxs_bulk, (Nx,Ny))
    v_dxdX    = rview(dxdX, idxs_bulk, (Nx,Ny))
    v_dydY    = rview(dydY, idxs_bulk, (Nx,Ny))

    # faces
    idxs_lhs, idxs_rhs, idxs_down, idxs_up = faceindices(mesh, kk)
    v_nf_lhs  = view(nf, idxs_lhs)
    v_nf_rhs  = view(nf, idxs_rhs)
    v_nf_down = view(nf, idxs_down)
    v_nf_up   = view(nf, idxs_up)

    # atm we can utilize that cartesian grids have constant jacobians and
    # pull those out of the loop
    dxdXij = v_dxdX[1,1]
    dydYij = v_dydY[1,1]

    # notes on @turbo:
    # - can't insert x_rhsij *= dxdXij after the k loop,
    #   cf. https://github.com/JuliaSIMD/LoopVectorization.jl/issues/506
    # - can't use k as index for both inner loops, otherwise gives wrong result
    #   in fact LV can't have more than one loop nested on the same level, cf. further
    #   discussion in issues/506; luckily we can split of one of the matrix products here
    #   and use a mul! instead
    # - for 4x4 cells @turbo is 2x faster than mul!, because latter allocates
    @turbo for j in 1:Ny
      for i in 1:Nx
        x_rhsij = 0.0
        for k in 1:Nx
          x_rhsij += MDM[i,k] * v_fx[k,j]
        end
        v_rhs[i,j] = x_rhsij * dxdXij
      end
    end
    @turbo for j in 1:Ny
      for i in 1:Nx
        y_rhsij = 0.0
        for l in 1:Ny
          y_rhsij += MDM[j,l] * v_fy[i,l]
        end
        v_rhs[i,j] += y_rhsij * dydYij
      end
    end
    # mul!(v_rhs, MDM, v_fx)
    # mul!(v_rhs, v_fy, transpose(MDM), dydYij, dxdXij)

    @turbo for j in 1:Ny
      for i in 1:Nx
        bdry_rhsij = (v_nf_rhs[j] * Ml_rhs[i] + v_nf_lhs[j]  * Ml_lhs[i])  * dxdXij +
                     (v_nf_up[i]  * Ml_up[j]  + v_nf_down[i] * Ml_down[j]) * dydYij
        v_rhs[i,j] -= bdry_rhsij
      end
    end

  end
  return
end

"""
    compute_rhs_weak_form!(rhs, fx, _::Real, nf, mesh::Mesh2d{DGElement})
    compute_rhs_weak_form!(rhs, _::Real, fy, nf, mesh::Mesh2d{DGElement})

Specialized version of `compute_rhs_weak_form!` where only the fluxs
`fx` or `fy` are applied, respectively.
"""
function compute_rhs_weak_form!(rhs, fx, _::Real, nf, mesh::Mesh2d{DGElement})
  @unpack element = mesh
  @unpack dxdX, dydY = get_static_variables(mesh.cache)
  @unpack invM, MDM, Ml_lhs, Ml_rhs, Npts = element
  Ml_down, Ml_up = Ml_lhs, Ml_rhs
  Nx, Ny = Npts, Npts
  Ncs = n_cells(mesh.tree)
  @inbounds for kk in 1:Ncs

    # bulk
    idxs_bulk = flatcellindices(mesh, kk)
    v_rhs     = rview(rhs,  idxs_bulk, (Nx,Ny))
    v_fx      = rview(fx,   idxs_bulk, (Nx,Ny))
    v_dxdX    = rview(dxdX, idxs_bulk, (Nx,Ny))
    v_dydY    = rview(dydY, idxs_bulk, (Nx,Ny))

    # faces
    idxs_lhs, idxs_rhs, idxs_down, idxs_up = faceindices(mesh, kk)

    v_nf_lhs  = view(nf, idxs_lhs)
    v_nf_rhs  = view(nf, idxs_rhs)
    v_nf_down = view(nf, idxs_down)
    v_nf_up   = view(nf, idxs_up)

    # atm we can utilize that cartesian grids have constant jacobians and
    # pull those out of the loop
    dxdXij = v_dxdX[1,1]
    dydYij = v_dydY[1,1]

    # for 4x4 cells @turbo is 2x faster than mul!, because latter allocates
    @turbo for j in 1:Ny
      for i in 1:Nx
        x_rhsij = 0.0
        for k in 1:Nx
          x_rhsij += MDM[i,k] * v_fx[k,j]
        end
        # note adding Jacobian factor in loop below
        v_rhs[i,j] = x_rhsij
      end
    end
    # mul!(v_rhs, MDM, v_fx)

    @turbo for j in 1:Ny
      for i in 1:Nx
        bdry_rhsij = (v_nf_rhs[j] * Ml_rhs[i] + v_nf_lhs[j]  * Ml_lhs[i]) * dxdXij
        # TODO Can we really neglect num fluxes in orthogonal direction in general?
        #      Or we need to adjust for triangular meshes?
        # bdry_rhsij -= (v_nf_rhs[j] * Ml_rhs[i] + v_nf_lhs[j]  * Ml_lhs[i])  * dxdXij +
        #               (v_nf_up[i]  * Ml_up[j]  + v_nf_down[i] * Ml_down[j]) * dydYij
        v_rhs[i,j] = v_rhs[i,j] * dxdXij - bdry_rhsij
      end
    end

  end
  return
end
function compute_rhs_weak_form!(rhs, _::Real, fy, nf, mesh::Mesh2d{DGElement})
  @unpack element = mesh
  @unpack dxdX, dydY = get_static_variables(mesh.cache)
  @unpack invM, MDM, Ml_lhs, Ml_rhs, Npts = element
  Ml_down, Ml_up = Ml_lhs, Ml_rhs
  Nx, Ny = Npts, Npts
  Ncs = n_cells(mesh.tree)
  @inbounds for kk in 1:Ncs

    # bulk
    idxs_bulk = flatcellindices(mesh, kk)
    v_rhs     = rview(rhs,  idxs_bulk, (Nx,Ny))
    v_fy      = rview(fy,   idxs_bulk, (Nx,Ny))
    v_dxdX    = rview(dydY, idxs_bulk, (Nx,Ny))
    v_dydY    = rview(dydY, idxs_bulk, (Nx,Ny))

    # faces
    idxs_lhs, idxs_rhs, idxs_down, idxs_up = faceindices(mesh, kk)

    v_nf_lhs  = view(nf, idxs_lhs)
    v_nf_rhs  = view(nf, idxs_rhs)
    v_nf_down = view(nf, idxs_down)
    v_nf_up   = view(nf, idxs_up)

    # atm we can utilize that cartesian grids have constant jacobians and
    # pull those out of the loop
    dxdXij = v_dxdX[1,1]
    dydYij = v_dydY[1,1]

    # for 4x4 cells @turbo is 2x faster than mul!, because latter allocates
    @turbo for j in 1:Ny
      for i in 1:Nx
        y_rhsij = 0.0
        for l in 1:Ny
          y_rhsij += MDM[j,l] * v_fy[i,l]
        end
        # note adding Jacobian factor in loop below
        v_rhs[i,j] = y_rhsij
      end
    end
    # mul!(v_rhs, v_fy, transpose(MDM))

    @turbo for j in 1:Ny
      for i in 1:Nx
        bdry_rhsij = (v_nf_up[i]  * Ml_up[j]  + v_nf_down[i] * Ml_down[j]) * dydYij
        # TODO Can we really neglect num fluxes in orthogonal direction in general?
        #      Or we need to adjust for triangular meshes?
        # bdry_rhsij -= (v_nf_rhs[j] * Ml_rhs[i] + v_nf_lhs[j]  * Ml_lhs[i])  * dxdXij +
        #               (v_nf_up[i]  * Ml_up[j]  + v_nf_down[i] * Ml_down[j]) * dydYij
        v_rhs[i,j] = v_rhs[i,j] * dydYij - bdry_rhsij
      end
    end

  end
  return
end
function compute_rhs_weak_form!(rhs, fx, fy, s, nf, mesh::Mesh2d{DGElement})
  compute_rhs_weak_form!(rhs, fx, fy, nf, mesh)
  rhs .+= s
  return
end
